#ifndef RPARSER_H
#define RPARSER_H

#include <vector>
#include <string>

class Token {
public:
    enum {
        NUMBER = 0,
        NAME = 1,
        PLUS = 2, 
        MINUS = 3,
        MUL = 4,
        DIV = 5,
        POW = 6,
        LPAR = 7,
        RPAR = 8,
        CM = 9,
        X = 10,
        Y = 11,
        Z = 12,
        ENDLINE = 13,
        ENDSTREAM = 14,
        UNDEFINED = 15
    };        
    int token;
    std::string text;
    double value;
    double (*func)(double);
public:
    Token(
        int t = UNDEFINED,
        const std::string& txt = std::string(), 
        double val = 0., 
        double (*f)(double) = 0
    ):
        token(t),
        text(txt),
        value(val),
        func(f)
    {}

    const char* tokenName() const;
    std::string toString() const;
};

class RParser {
public:
    std::vector<Token> rpn;
    std::string input;
    std::string errorText;
    std::vector<Token> tokens;
    int tokenPos;
public:
    RParser():
        rpn(),
        input(),
        errorText(),
        tokens(),
        tokenPos(0)
    {}
    
    void initialize(const std::string& str);
    bool parse();
    double evaluate(double x = 0., double y = 0., double z = 0.) const;
private:
    /*
    Grammar rules:
    S -> F $        // Parsing text is a Formula with end of line marker
    F -> T FTail    // Formula is a Term and a tail of Formula
    FTail -> eps | + T FTail | - T FTail  // Tail of Formula
    T -> M TTail    // Term is a Multiplier and a tail of Term
    TTail -> eps | * M TTail | / M TTail  // Tail of Term
    M -> E MTail    // Multiplier is Elementary formula and a tail of Multiplier
    MTail -> eps | ^ M  // Tail of Multiplier
    E -> number | x | y | z | func | ( F )  // Elementary formula
    Func -> name ( Args )     // Function call
    Args -> eps | ArgList     // Args of function (either empty of non-empty list
    ArgList -> F ArgListTail  // ArgList is a Formula and a tail of arg. list
    ArgListTail -> eps | , F ArgListTail    // Tail of argument list
    */
    
    void processF();
    void processFTail();
    void processT();
    void processTTail();
    void processM();
    void processMTail();
    void processE();
    void processFunc();
    void processArgs();
    void processArgList();
    void processArgListTail();
private:
    // Scanner part
    Token peekToken();
    void skipToken(int tokenType = (-1));
    void popToken();
};    

#endif
